"""
.. module:: read_slab
   :synopsis: Read routines and data structure for prodimo slab models.

.. moduleauthor:: A. M. Arabhavi


"""

import os
import string

from astropy.convolution import Gaussian1DKernel
from astropy.convolution import convolve as apy_convolve
from scipy.constants import h,c,k
from scipy.special import erf as serf

import numpy as np
import pandas as pd


class slab_data:
    """
    Structure of the models set up from the SlabInput.in file:

    ProDiMo run

    |__Model_1
        |__molecule_1
        |__molecule_2
        |__...
        |__molecule_n
    |__Model_2
        |__molecule_1
        |__molecule_2
        |__...
        |__molecule_n
    |__...
        |__...
        |__...
        |__...
    |__Model_n
        |__molecule_1
        |__molecule_2
        |__...
        |__molecule_n

    Data structure to hold the corresponding slab model output
    slab_data

    |_directory

    |_models
        |__model_1 (type:slab)
           |__prodimo_Model_1_molecule_1
        |__model_2 (type:slab)
           |__prodimo_Model_1_molecule_2
        |__model_... (type:slab)
           |__...
        |__model_n (type:slab)
           |__prodimo_Model_2_molecule_n
        |__model_n+1 (type:slab)
           |__prodimo_Model_2_molecule_1
        |__model_n+2 (type:slab)
           |__prodimo_Model_2_molecule_2
        |__model_... (type:slab)
           |__...
        |__model_n+n (type:slab)
           |__prodimo_model_2_molecule_n
        |__model_... (type:slab)
           |__...
    """

    def __init__(self):
        self._nmodels=0
        """ integer :
        number of models.
        """

        self.directory=None
        """ string :
        The directory from which the model was read.
        Can be a relative path.
        """
        self.models=[]
        """ list :
        The list containing the models
        """

        self._species_number=None
        """ list:
        It stores the species number according to SlabInput.in of all models
        """

        self._model_number=None
        """ list:
        It stores the model number according to SlabInput.in of all models
        """

        self._NH=None
        """ list:
        It stores the total gas densities of all models
        """

        self._nColl=None
        """ list:
        It stores the total collision partner abundance of all models
        """

        self._ne=None
        """ list:
        It stores the electron abundance of all models
        """

        self._nHe=None
        """ list:
        It stores the He abundance of all models
        """

        self._nHII=None
        """ list:
        It stores the HII abundance of all models
        """

        self._nHI=None
        """ list:
        It stores the HI abundance of all models
        """

        self._nH2=None
        """ list:
        It stores the molecular hydrogen abundance of all models
        """

        self._dust_to_gas=None
        """ list:
        It stores the dust to gas ratio of all models
        """

        self._vturb=None
        """ list:
        It stores the turbulent broadening of all models
        """

        self._Tg=None
        """ list:
        It stores the gas temperatures of all models
        """

        self._Td=None
        """ list:
        It stores the dust temperatures of all models
        """

        self._species_name=None
        """ list:
        It stores the name of the molecules of all models
        """

        self._species_index=None
        """ list:
        It stores the ProDiMo species index of all models
        """

        self._abundance=None
        """ list:
        It stores the species abundance according to SlabInput.in of all models
        """

        self._dv=None
        """ list:
        It stores the velocity width of all models
        """

        self._nlevels=None
        """ list:
        It stores the number of levels of all models
        """

        self._nlines=None
        """ list:
        It stores the number of lines of all models
        """
        self._linedata=None
        """ list:
        It stores the line data of all models
        """

        self._leveldata=None
        """ list:
        It stores the level data of all models
        """

        self._convWavelength=None
        """ list:
        It stores the convolved wavelength grid of all models
        """

        self._convLTEflux=None
        """ list:
        It stores the convolved LTE flux of all models
        """

        self._convNLTEflux=None
        """ list:
        It stores the convolved NLTE flux of all models
        """

        self._convType=None
        """ list:
        It stores the convolution type of all models
        """

        self._convR=None
        """ list:
        It stores the convolved resolving power R of all models
        """

    def __str__(self):
        output="Info: \n"
        output+="\n NModels: "
        output+=str(self.nmodels)
        output+="\n\n Directory: "+self.directory
        return output

    def show(self):
        print(self)
        for i in range(self.nmodels):
            self.models[i].show()
            print('\n')

    def add_model(self,model_data):
        if not isinstance(model_data,slab):
            raise TypeError('path must be of class:model')
        self.models.append(model_data)

    def remove_model(self,index):
        if index<0:
            index+=self.nmodels
        if index+1>self.nmodels or index<0:
            raise IndexError('Index out of range')
        else:
            self.nmodels-=1
            self.models.pop(index)

    def __getitem__(self,arg):
        ret_data=slab_data()
        if isinstance(arg,int):
            if arg>self.nmodels:
                raise ValueError(f'index {arg} greater than number of models {self.nmodels}')
            ret_data.add_model(self.models[arg])
        elif isinstance(arg,type(None)):
            for model in self.models:
                ret_data.add_model(model)
        elif isinstance(arg,slice):
            if isinstance(arg.start,int):
                if isinstance(arg.stop,int):
                    for model in self.models[arg]:
                        ret_data.add_model(model)
                elif isinstance(arg.stop,type(None)):
                    ret_data.add_model(self.models[arg.start])
                else:
                    raise IndexError(f'slicing not understandable {arg}')
            elif isinstance(arg.start,str):
                if arg.start=='species_name':
                    sel_species=[]
                    if isinstance(arg.stop,str):
                        sel_species.append(arg.stop)
                    elif isinstance(arg.stop,list):
                        sel_species=arg.stop
                    if len(sel_species)>0:
                        for i in range(self.nmodels):
                            if self.models[i].species_name in sel_species:
                                ret_data.add_model(self.models[i])
                elif arg.start=='species_number':
                    sel_species_number=[]
                    if isinstance(arg.stop,int):
                        if isinstance(arg.step,int):
                            for i in range(arg.stop,arg.step):
                                sel_species_number.append(i)
                        else:
                            sel_species_number.append(arg.stop)
                    elif isinstance(arg.stop,list):
                            sel_species_number=arg.step
                    if len(sel_species_number)>0:
                        for i in range(self.nmodels):
                            if self.models[i].species_number in sel_species_number:
                                ret_data.add_model(self.models[i])
                elif arg.start=='model_number':
                    sel_species_number=[]
                    if isinstance(arg.stop,int):
                        if isinstance(arg.step,int):
                            for i in range(arg.stop,arg.step):
                                sel_species_number.append(i)
                        else:
                            sel_species_number.append(arg.stop)
                    elif isinstance(arg.stop,list):
                            sel_species_number=arg.step
                    if len(sel_species_number)>0:
                        for i in range(self.nmodels):
                            if self.models[i].model_number in sel_species_number:
                                ret_data.add_model(self.models[i])
                elif arg.start=='NH':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].NH<upper and self.models[i].NH>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].NH==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].NH in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='nColl':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].nColl<upper and self.models[i].nColl>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].nColl==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].nColl in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='ne':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].ne<upper and self.models[i].ne>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].ne==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):

                        for i in range(self.nmodels):
                            if self.models[i].ne in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='nHe':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].nHe<upper and self.models[i].nHe>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].nHe==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):

                        for i in range(self.nmodels):
                            if self.models[i].nHe in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='nHII':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].nHII<upper and self.models[i].nHII>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].nHII==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].nHII in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='nHI':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].nHI<upper and self.models[i].nHI>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].nHI==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].nHI in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='nH2':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].nH2<upper and self.models[i].nH2>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].nH2==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):

                        for i in range(self.nmodels):
                            if self.models[i].nH2 in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='dust_to_gas':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].dust_to_gas<upper and self.models[i].dust_to_gas>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].dust_to_gas==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):

                        for i in range(self.nmodels):
                            if self.models[i].dust_to_gas in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='vturb':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].vturb<upper and self.models[i].vturb>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].vturb==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].vturb in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='Tg':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].Tg<upper and self.models[i].Tg>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].Tg==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].Tg in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='Td':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].Td<upper and self.models[i].Td>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].Td==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].Td in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='species_index':
                    sel_species_number=[]
                    if isinstance(arg.stop,int):
                        if isinstance(arg.step,int):
                            for i in range(arg.stop,arg.step):
                                sel_species_number.append(i)
                        else:
                            sel_species_number.append(arg.stop)
                    elif isinstance(arg.stop,list):
                            sel_species_number=arg.step
                    if len(sel_species_number)>0:
                        for i in range(self.nmodels):
                            if self.models[i].species_index in sel_species_number:
                                ret_data.add_model(self.models[i])
                elif arg.start=='abundance':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].abundance<upper and self.models[i].abundance>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].abundance==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].abundance in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='dv':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].dv<upper and self.models[i].dv>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].dv==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].dv in arg.step:
                                ret_data.add_model(self.models[i])
                elif arg.start=='nlevels':
                    sel_species_number=[]
                    if isinstance(arg.stop,int):
                        if isinstance(arg.step,int):
                            for i in range(arg.stop,arg.step):
                                sel_species_number.append(i)
                        else:
                            sel_species_number.append(arg.stop)
                    elif isinstance(arg.stop,list):
                            sel_species_number=arg.step
                    if len(sel_species_number)>0:
                        for i in range(self.nmodels):
                            if self.models[i].nlevels in sel_species_number:
                                ret_data.add_model(self.models[i])
                elif arg.start=='nlines':
                    sel_species_number=[]
                    if isinstance(arg.stop,int):
                        if isinstance(arg.step,int):
                            for i in range(arg.stop,arg.step):
                                sel_species_number.append(i)
                        else:
                            sel_species_number.append(arg.stop)
                    elif isinstance(arg.stop,list):
                            sel_species_number=arg.step
                    if len(sel_species_number)>0:
                        for i in range(self.nmodels):
                            if self.models[i].nlines in sel_species_number:
                                ret_data.add_model(self.models[i])
                elif arg.start=='convType':
                    sel_species_number=[]
                    if isinstance(arg.stop,int):
                        if isinstance(arg.step,int):
                            sel_species_number=[arg.stop,arg.step]
                        else:
                            sel_species_number.append(arg.stop)
                    elif isinstance(arg.stop,list):
                            sel_species_number=arg.step
                    if len(sel_species_number)>0:
                        for i in range(self.nmodels):
                            if self.models[i].convType in sel_species_number:
                                ret_data.add_model(self.models[i])
                elif arg.start=='convR':
                    upper=np.nan
                    lower=np.nan
                    if isinstance(arg.stop,float) or isinstance(arg.stop,int):
                        if isinstance(arg.step,float) or isinstance(arg.step,int):
                            lower,upper=arg.stop,arg.step
                        else:
                            upper=arg.stop
                        for i in range(self.nmodels):
                            if not np.isnan(lower):
                                if self.models[i].convR<upper and self.models[i].convR>lower:
                                    ret_data.add_model(self.models[i])
                            else:
                                if self.models[i].convR==upper:
                                    ret_data.add_model(self.models[i])
                    elif isinstance(arg.stop,list):
                        for i in range(self.nmodels):
                            if self.models[i].convR in arg.step:
                                ret_data.add_model(self.models[i])
                else:
                    raise KeyError('Unrecognised parameter for slicing')
        return ret_data

    def convolve(self,freq_bins=None,R=1,lambda_0=1,lambda_n=1,vr=1300,NLTE=False,conv_type=1,verbose=True):
        for i,d in enumerate(self.models):
            if verbose: print('\n',i+1,'; Model number ',self.models[i].model_number,'; Species number ',self.models[i].species_number)
            d.convolve(freq_bins=freq_bins,R=R,lambda_0=lambda_0,lambda_n=lambda_n,vr=vr,NLTE=NLTE,conv_type=conv_type)

    @property
    def nmodels(self):
        self._nmodels=len(self.models)
        return self._nmodels

    @nmodels.setter
    def nmodels(self,value):
        self._nmodels=value

    @property
    def species_number(self):
        if self.nmodels>1:
            self._species_number=[]
            for i in range(self.nmodels):
                self._species_number.append(self.models[i].species_number)
        else:
            self._species_number=self.models[0].species_number
        return self._species_number

    @species_number.setter
    def species_number(self,value):
        self._species_number=value

    @property
    def model_number(self):
        if self.nmodels>1:
            self._model_number=[]
            for i in range(self.nmodels):
                self._model_number.append(self.models[i].model_number)
        else:
            self._model_number=self.models[0].model_number
        return self._model_number

    @model_number.setter
    def model_number(self,value):
        self._model_number=value

    @property
    def NH(self):
        if self.nmodels>1:
            self._NH=[]
            for i in range(self.nmodels):
                self._NH.append(self.models[i].NH)
        else:
            self._NH=self.models[0].NH
        return self._NH

    @NH.setter
    def NH(self,value):
        self._NH=value

    @property
    def nColl(self):
        if self.nmodels>1:
            self._nColl=[]
            for i in range(self.nmodels):
                self._nColl.append(self.models[i].nColl)
        else:
            self._nColl=self.models[0].nColl
        return self._nColl

    @nColl.setter
    def nColl(self,value):
        self._nColl=value

    @property
    def ne(self):
        if self.nmodels>1:
            self._ne=[]
            for i in range(self.nmodels):
                self._ne.append(self.models[i].ne)
        else:
            self._ne=self.models[0].ne
        return self._ne

    @ne.setter
    def ne(self,value):
        self._ne=value

    @property
    def nHe(self):
        if self.nmodels>1:
            self._nHe=[]
            for i in range(self.nmodels):
                self._nHe.append(self.models[i].nHe)
        else:
            self._nHe=self.models[0].nHe
        return self._nHe

    @nHe.setter
    def nHe(self,value):
        self._nHe=value

    @property
    def nHII(self):
        if self.nmodels>1:
            self._nHII=[]
            for i in range(self.nmodels):
                self._nHII.append(self.models[i].nHII)
        else:
            self._nHII=self.models[0].nHII
        return self._nHII

    @nHII.setter
    def nHII(self,value):
        self._nHII=value

    @property
    def nHI(self):
        if self.nmodels>1:
            self._nHI=[]
            for i in range(self.nmodels):
                self._nHI.append(self.models[i].nHI)
        else:
            self._nHI=self.models[0].nHI
        return self._nHI

    @nHI.setter
    def nHI(self,value):
        self._nHI=value

    @property
    def nH2(self):
        if self.nmodels>1:
            self._nH2=[]
            for i in range(self.nmodels):
                self._nH2.append(self.models[i].nH2)
        else:
            self._nH2=self.models[0].nH2
        return self._nH2

    @nH2.setter
    def nH2(self,value):
        self._nH2=value

    @property
    def dust_to_gas(self):
        if self.nmodels>1:
            self._dust_to_gas=[]
            for i in range(self.nmodels):
                self._dust_to_gas.append(self.models[i].dust_to_gas)
        else:
            self._dust_to_gas=self.models[0].dust_to_gas
        return self._dust_to_gas

    @dust_to_gas.setter
    def dust_to_gas(self,value):
        self._dust_to_gas=value

    @property
    def vturb(self):
        if self.nmodels>1:
            self._vturb=[]
            for i in range(self.nmodels):
                self._vturb.append(self.models[i].vturb)
        else:
            self._vturb=self.models[0].vturb
        return self._vturb

    @vturb.setter
    def vturb(self,value):
        self._vturb=value

    @property
    def Tg(self):
        if self.nmodels>1:
            self._Tg=[]
            for i in range(self.nmodels):
                self._Tg.append(self.models[i].Tg)
        else:
            self._Tg=self.models[0].Tg
        return self._Tg

    @Tg.setter
    def Tg(self,value):
        self._Tg=value

    @property
    def Td(self):
        if self.nmodels>1:
            self._Td=[]
            for i in range(self.nmodels):
                self._Td.append(self.models[i].Td)
        else:
            self._Td=self.models[0].Td
        return self._Td

    @Td.setter
    def Td(self,value):
        self._Td=value

    @property
    def species_name(self):
        if self.nmodels>1:
            self._species_name=[]
            for i in range(self.nmodels):
                self._species_name.append(self.models[i].species_name)
        else:
            self._species_name=self.models[0].species_name
        return self._species_name

    @species_name.setter
    def species_name(self,value):
        self._species_name=value

    @property
    def species_index(self):
        if self.nmodels>1:
            self._species_index=[]
            for i in range(self.nmodels):
                self._species_index.append(self.models[i].species_index)
        else:
            self._species_index=self.models[0].species_index
        return self._species_index

    @species_index.setter
    def species_index(self,value):
        self._species_index=value

    @property
    def abundance(self):
        if self.nmodels>1:
            self._abundance=[]
            for i in range(self.nmodels):
                self._abundance.append(self.models[i].abundance)
        else:
            self._abundance=self.models[0].abundance
        return self._abundance

    @abundance.setter
    def abundance(self,value):
        self._abundance=value

    @property
    def dv(self):
        if self.nmodels>1:
            self._dv=[]
            for i in range(self.nmodels):
                self._dv.append(self.models[i].dv)
        else:
            self._dv=self.models[0].dv
        return self._dv

    @dv.setter
    def dv(self,value):
        self._dv=value

    @property
    def nlevels(self):
        if self.nmodels>1:
            self._nlevels=[]
            for i in range(self.nmodels):
                self._nlevels.append(self.models[i].nlevels)
        else:
            self._nlevels=self.models[0].nlevels
        return self._nlevels

    @nlevels.setter
    def nlevels(self,value):
        self._nlevels=value

    @property
    def nlines(self):
        if self.nmodels>1:
            self._nlines=[]
            for i in range(self.nmodels):
                self._nlines.append(self.models[i].nlines)
        else:
            self._nlines=self.models[0].nlines
        return self._nlines

    @nlines.setter
    def nlines(self,value):
        self._nlines=value

    @property
    def linedata(self):
        if self.nmodels>1:
            self._linedata=[]
            for i in range(self.nmodels):
                self._linedata.append(self.models[i].linedata)
        else:
            self._linedata=self.models[0].linedata
        return self._linedata

    @linedata.setter
    def linedata(self,value):
        self._linedata=value

    @property
    def leveldata(self):
        if self.nmodels>1:
            self._leveldata=[]
            for i in range(self.nmodels):
                self._leveldata.append(self.models[i].leveldata)
        else:
            self._leveldata=self.models[0].leveldata
        return self._leveldata

    @leveldata.setter
    def leveldata(self,value):
        self._leveldata=value

    @property
    def convWavelength(self):
        if self.nmodels>1:
            self._convWavelength=[]
            for i in range(self.nmodels):
                self._convWavelength.append(self.models[i].convWavelength)
        else:
            self._convWavelength=self.models[0].convWavelength
        return self._convWavelength

    @convWavelength.setter
    def convWavelength(self,value):
        self._convWavelength=value

    @property
    def convLTEflux(self):
        if self.nmodels>1:
            self._convLTEflux=[]
            for i in range(self.nmodels):
                self._convLTEflux.append(self.models[i].convLTEflux)
        else:
            self._convLTEflux=self.models[0].convLTEflux
        return self._convLTEflux

    @convLTEflux.setter
    def convLTEflux(self,value):
        self._convLTEflux=value

    @property
    def convNLTEflux(self):
        if self.nmodels>1:
            self._convNLTEflux=[]
            for i in range(self.nmodels):
                self._convNLTEflux.append(self.models[i].convNLTEflux)
        else:
            self._convNLTEflux=self.models[0].convNLTEflux
        return self._convNLTEflux

    @convNLTEflux.setter
    def convNLTEflux(self,value):
        self._convNLTEflux=value

    @property
    def convType(self):
        if self.nmodels>1:
            self._convType=[]
            for i in range(self.nmodels):
                self._convType.append(self.models[i].convType)
        else:
            self._convType=self.models[0].convType
        return self._convType

    @convType.setter
    def convType(self,value):
        self._convType=value

    @property
    def convR(self):
        if self.nmodels>1:
            self._convR=[]
            for i in range(self.nmodels):
                self._convR.append(self.models[i].convR)
        else:
            self._convR=self.models[0].convR
        return self._convR

    @convR.setter
    def convR(self,value):
        self._convR=value


class slab:
    """
    class:: slab
    Class to hold the data of individual slab models
    """

    def __init__(self):
        self.species_number=None
        self.model_number=None
        self.NH=None
        self.nColl=None
        self.ne=None
        self.nHe=None
        self.nHII=None
        self.nHI=None
        self.nH2=None
        self.dust_to_gas=None
        self.vturb=None
        self.Tg=None
        self.Td=None
        self.species_name=None
        self.species_index=None
        self.abundance=None
        self.dv=None
        self.nlevels=None
        self.nlines=None
        self.linedata=None
        self.leveldata=None
        self.convWavelength=None
        self.convLTEflux=None
        self.convNLTEflux=None
        self.convType=None
        self.convR=None

    def __str__(self):
        output="Info Model: \n"
        output+='species_number= '
        output+=str(self.species_number)+'\n\n'
        output+='model_number  = '
        output+=str(self.model_number)+'\n\n'
        output+='NH            = '
        output+=str(self.NH)+'\n\n'
        output+='nColl         = '
        output+=str(self.nColl)+'\n\n'
        output+='ne            = '
        output+=str(self.ne)+'\n\n'
        output+='nHe           = '
        output+=str(self.nHe)+'\n\n'
        output+='nHII          = '
        output+=str(self.nHII)+'\n\n'
        output+='nHI           = '
        output+=str(self.nHI)+'\n\n'
        output+='nH2           = '
        output+=str(self.nH2)+'\n\n'
        output+='dust_to_gas   = '
        output+=str(self.dust_to_gas)+'\n\n'
        output+='vturb         = '
        output+=str(self.vturb)+'\n\n'
        output+='Tg            = '
        output+=str(self.Tg)+'\n\n'
        output+='Td            = '
        output+=str(self.Td)+'\n\n'
        output+='species_name   = '
        output+=self.species_name+'\n\n'
        output+='abundance     = '
        output+=str(self.abundance)+'\n\n'
        output+='dv            = '
        output+=str(self.dv)+'\n\n'
        output+='nlevels       = '
        output+=str(self.nlevels)+'\n\n'
        output+='nlines        = '
        output+=str(self.nlines)+'\n\n'
        output+='convType       = '
        output+=str(self.convType)+'\n\n'
        output+='convR        = '
        output+=str(self.convR)+'\n\n'
        return output

    def convolve(self,freq_bins=None,R=1,lambda_0=1,lambda_n=1,vr=1300,NLTE=False,conv_type=1):
        t='FLTE'
        if NLTE:
            t='FNLTE'
        da=self.linedata
        if lambda_0==lambda_n:
            lambda_0=np.amin(da['GHz']>c/lambda_n*1e-3)
            lambda_n=np.amax(da['GHz']>c/lambda_n*1e-3)
            lambda_0=lambda_0*10**-0.05
            lambda_n=lambda_n*10**0.05
        da_req=da[(da['GHz']>c/lambda_n*1e-3)&(da['GHz']<c/lambda_0*1e-3)]
        da_req.reset_index(drop=True,inplace=True)
        if conv_type==1:
            l,f=generalConvolve(da_req[t],da_req['GHz'],R=R,lambda_0=lambda_0,lambda_n=lambda_n)
        else:
            l,f=specConvolve(da_req[t],da_req['GHz'],freq_bins=freq_bins,R=R,lambda_0=lambda_0,lambda_n=lambda_n,vr=vr)
        self.convWavelength=l
        self.convType=conv_type
        self.convR=R
        if NLTE:
            self.convNLTEflux=f
        else:
            self.convLTEflux=f

    def show(self):
        print(self)
        print('lineData    = ')
        print(self.linedata)
        print('levelData   = ')
        print(self.leveldata)


def read_slab(model_path='SlabResults.out',verbose=True):
    """
    Function to read slab model output
    """
    if isinstance(model_path,list):
        data=slab_data()
        data.directory=model_path
        for i in model_path:
            rdata=read_slab(i,verbose=verbose)
            for j in range(rdata.nmodels):
                data.add_model(rdata.models[j])
        return(data)

    if verbose: print("Reading slab model output from: ",model_path)
    f=open(model_path)
    data=slab_data()
    nmodels=int(f.readline().split()[0])
    data.directory=model_path
    for i in range(nmodels):
#         tempo                =   slab()
        tempo_model_number=int(f.readline().split()[0])
        nspecies=int(f.readline().split()[0])
#         tempo.species_number =   int(f.readline().split()[0])
        tempo_nHtot=float(f.readline().split()[0])
        tempo_gasDens=float(f.readline().split()[0])
        tempo_nelec=float(f.readline().split()[0])
        tempo_nHe=float(f.readline().split()[0])
        tempo_nHII=float(f.readline().split()[0])
        tempo_nHI=float(f.readline().split()[0])
        tempo_nH2=float(f.readline().split()[0])
        tempo_dust_to_gas=float(f.readline().split()[0])
        tempo_vturb=float(f.readline().split()[0])
        tempo_Tg=float(f.readline().split()[0])
        tempo_Td=float(f.readline().split()[0])

        for j in range(nspecies):
            tempo=slab()
            tempo.model_number=tempo_model_number
            tempo.NH=tempo_nHtot
            tempo.nColl=tempo_gasDens
            tempo.ne=tempo_nelec
            tempo.nHe=tempo_nHe
            tempo.nHII=tempo_nHII
            tempo.nHI=tempo_nHI
            tempo.nH2=tempo_nH2
            tempo.dust_to_gas=tempo_dust_to_gas
            tempo.vturb=tempo_vturb
            tempo.Tg=tempo_Tg
            tempo.Td=tempo_Td

            tempo.species_index=int(f.readline().split()[0])
            tempo.species_number=j+1
            tempo.species_name=f.readline().split()[0]
            tempo.abundance=float(f.readline().split()[0])
            tempo.dv=float(f.readline().split()[0])
            tempo.nlevels=int(f.readline().split()[0])
            lines=f.readline()

            tempo.nlines=None
            tempo.linedata=None
            tempo.leveldata=None
            tempo.convWavelength=None
            tempo.convLTEflux=None
            tempo.convNLTEflux=None
            tempo.convType=None
            tempo.convR=None

            dat=[np.zeros((tempo.nlevels,8)),[],[]]
            for k in range(tempo.nlevels):
                dat[0][k,:]=np.asarray([float(x) for x in f.readline().split()])
            tempo.nlines=int(f.readline().split()[0])
            lines=f.readline()
            lines=f.readline()
            dat[1]=np.zeros((tempo.nlines,23))
            slabOutFormat=[9,9,9,5,5,5,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,21,21,21,21]
            for k in range(tempo.nlines):
                lineRead=[]
                lines=f.readline()
                l=0
                for length in slabOutFormat:
                    lineRead.append(lines[l:l+length])
                    l+=length
                dat[1][k,:]=np.asarray([float(x) for x in lineRead[0:-4]])
                dat[2].append([a.strip() for a in lineRead[-4:]])

            lineData=pd.DataFrame(dat[1],columns=['i','u','l','e','v','J','gu','Eu','A','B','GHz','tauD','Jback','pop','ltepop','tauNLTE','tauLTE','bNLTE','bLTE','pNLTE','pLTE','FNLTE','FLTE'])
            lineData['global_u']=[a[0] for a in dat[2]]
            lineData['global_l']=[a[1] for a in dat[2]]
            lineData['local_u']=[a[2] for a in dat[2]]
            lineData['local_l']=[a[3] for a in dat[2]]
            levData=pd.DataFrame(dat[0],columns=['i','g','E','pop','ltepop','e','v','J'])
            tempo.linedata=lineData
            tempo.leveldata=levData
            data.add_model(tempo)
    return(data)


def specConvolve(lineStrengths,lineFreq,freq_bins=None,R=1,lambda_0=1,lambda_n=1,vr=1300):
    gPerc=lambda x,m,s: serf((x-m)/s)*0.5+0.5
    nu_ij=lineFreq*1e9
    if lambda_n==lambda_0:
        lambda_n=np.amax(c/nu_ij*1e6)*2
    if type(freq_bins)==type(None):
        nu=generate_grid(R=R,lambda_0=lambda_0,lambda_n=lambda_n)*1e9
    else:
        freq_bins=np.concatenate((freq_bins,np.array([(freq_bins[-1]-freq_bins[-2])+freq_bins[-1]])))
        nu=freq_bins*1e9
    nu_ij_tile=np.tile(nu_ij,[len(nu),1])
    nu_tile=np.tile(nu,[len(nu_ij),1])
    lineStrengths_tile=np.tile(lineStrengths,[len(nu),1])
    perc_tile=gPerc(nu_tile,nu_ij_tile.T,nu_ij_tile.T/c*vr)
    x=perc_tile[:,:-1]-perc_tile[:,1:]
    wavelength,flux=c/nu[:-1]*1e6,np.sum(x*lineStrengths_tile[:-1,:].T,axis=0)/(nu[:-1]-nu[1:])
    return wavelength,flux


def generalConvolve(lineStrengths,lineFreq,R=1,lambda_0=1,lambda_n=1):
    if lambda_n==lambda_0:
        lambda_n=c/np.amin(lineFreq)*1e-3*2

    R_back=1e6
    freq_bins=generate_grid(R_back,lambda_0,lambda_n)
    wave_bins=c/freq_bins*1e-3
    flux_bins=wave_bins*0.0
    i=6
    S=lineStrengths
    F=lineFreq
    ind=np.digitize(F,np.flip(freq_bins)[:-1])
    for i,idx in enumerate(ind):
        flux_bins[idx]+=S[i]

    FWHM=R_back/R/2.355
    g=Gaussian1DKernel(stddev=FWHM,factor=7)
    A=apy_convolve(flux_bins,g)
    A=A[1:]/(freq_bins[:-1]-freq_bins[1:])*1e-9
    W=np.flip(wave_bins[1:])
    return W,A


def generate_grid(R=1,lambda_0=1,lambda_n=1):
    del_loglam=np.log10(1.0+1.0/R)
    N=1+int(np.log10(lambda_n/lambda_0)/del_loglam)
    mwlsline=np.logspace(np.log10(lambda_0),np.log10(lambda_n),N)
    nu=c/mwlsline*1e6
    return(nu*1e-9)


def convolve(models,freq_bins=None,R=1,lambda_0=1,lambda_n=1,vr=1300,NLTE=False,conv_type=1,verbose=True):
    for i,d in enumerate(models):
        if verbose: print('\n\nModel ',i+1)
        d.convolve(freq_bins=freq_bins,R=R,lambda_0=lambda_0,lambda_n=lambda_n,vr=vr,NLTE=NLTE,conv_type=conv_type)
    return models


def generate_slab_grid(directory='.',grid_parameters={},combination=False,Ntot=None,Tg=None,nHplus=None,nH1=None,nH2=None,nHe=None,nelec=None,Td=None,vturb=None,dust_to_gas=None,species_list={},output_filename='SlabResults.out',separate_op_files=True):
    """
    Function to generate SlabInput.in input file containing the details of the slab model grid
    """
    nmodels=1
    combination=False
    if len(grid_parameters)>0:
        if not combination:
            nmodels=len(grid_parameters[list(grid_parameters.keys())[0]])
        else:
            for i in range(len(grid_parameters)):
                nmodels*=len(grid_parameters[list(grid_parameters.keys())[i]])

    Ntot_list=np.ones((nmodels))
    Tg_list=np.ones((nmodels))
    nHplus_list=np.ones((nmodels))
    nH1_list=np.ones((nmodels))
    nH2_list=np.ones((nmodels))
    nHe_list=np.ones((nmodels))
    nelec_list=np.ones((nmodels))
    Td_list=np.ones((nmodels))
    vturb_list=np.ones((nmodels))
    dust_to_gas_list=np.ones((nmodels))
    if len(species_list)<1: raise ValueError('The species_list cannot be empty')

    if not combination:
        for i in range(len(grid_parameters)):
            if list(grid_parameters.keys())[i]=='Ntot':
                Ntot_list*=grid_parameters['Ntot']
            elif list(grid_parameters.keys())[i]=='Tg':
                Tg_list*=grid_parameters['Tg']
            elif list(grid_parameters.keys())[i]=='nHplus':
                nHplus_list*=grid_parameters['nHplus']
            elif list(grid_parameters.keys())[i]=='nH1':
                nH1_list*=grid_parameters['nH1']
            elif list(grid_parameters.keys())[i]=='nH2':
                nH2_list*=grid_parameters['nH2']
            elif list(grid_parameters.keys())[i]=='nHe':
                nHe_list*=grid_parameters['nHe']
            elif list(grid_parameters.keys())[i]=='nelec':
                nelec_list*=grid_parameters['nelec']
            elif list(grid_parameters.keys())[i]=='Td':
                Td_list*=grid_parameters['Td']
            elif list(grid_parameters.keys())[i]=='vturb':
                vturb_list*=grid_parameters['vturb']
            elif list(grid_parameters.keys())[i]=='dust_to_gas':
                dust_to_gas_list*=grid_parameters['dust_to_gas']
            else: raise KeyError(f'Unidentified grid parameter {list(grid_parameters.keys())[i]}')
    else:
        print('Combination is still not supported')

    if np.sum(Ntot_list)==nmodels:
        if not (Ntot        is None):
            Ntot_list*=Ntot
        else:
            Ntot_list*=1e15
    if np.sum(Tg_list)==nmodels:
        if not (Tg          is None):
            Tg_list*=Tg
        else:
            Tg_list*=100
    if np.sum(nHplus_list)==nmodels:
        if not (nHplus      is None):
            nHplus_list*=nHplus
        else:
            nHplus_list*=1e1
    if np.sum(nH1_list)==nmodels:
        if not (nH1         is None):
            nH1_list*=nH1
        else:
            nH1_list*=1e12
    if np.sum(nH2_list)==nmodels:
        if not (nH2         is None):
            nH2_list*=nH2
        else:
            nH2_list*=1e12
    if np.sum(nHe_list)==nmodels:
        if not (nHe         is None):
            nHe_list*=nHe
        else:
            nHe_list*=1e11
    if np.sum(nelec_list)==nmodels:
        if not (nelec       is None):
            nelec_list*=nelec
        else:
            nelec_list*=1e8
    if np.sum(Td_list)==nmodels:
        if not (Td          is None):
            Td_list*=Td
        else:
            Td_list*=10
    if np.sum(vturb_list)==nmodels:
        if not (vturb       is None):
            vturb_list*=vturb
        else:
            vturb_list*=1.4
    if np.sum(dust_to_gas_list)==nmodels:
        if not (dust_to_gas is None):
            dust_to_gas_list*=dust_to_gas
        else:
            dust_to_gas_list*=1e-21

    os.system(f'touch {directory+"/SlabInput.in"}')
    f=open(directory+'/SlabInput.in','w')
    f.write("***********************************************************\n")
    f.write("*** Input file for slab escape probability with ProDiMo ***\n")
    f.write("***********************************************************\n")
    f.write("*** nmodels should follow output_filename ***\n")
    f.write(f'{output_filename} ! output_filename\n')
    if separate_op_files:
        f.write('.true.    ! separate_op_files\n')
    else:
        f.write('.false.   ! separate_op_files\n')
    f.write(f'{nmodels}             ! nmodels\n')
    f.write('-------------------------------------------------------')
    for i in range(nmodels):
        f.write(f'\n*** model {i+1} ***\n')
        f.write(fmt.format('{:.1m}',Ntot_list[i]))
        f.write('	   ! NHtot  [cm^-2] total gas column density\n')
        f.write(fmt.format('{:.1m}',nHplus_list[i]))
        f.write('	   ! nH+    [cm^-3]  proton number density\n')
        f.write(fmt.format('{:.1m}',nH1_list[i]))
        f.write('	   ! nHI	 [cm^-3] atomic hydrogen\n')
        f.write(fmt.format('{:.1m}',nH2_list[i]))
        f.write('	   ! nH2    [cm^-3]   molecular hydrogen\n')
        f.write(fmt.format('{:.1m}',nHe_list[i]))
        f.write('	   ! nHe    [cm^-3]   Helium\n')
        f.write(fmt.format('{:.1m}',nelec_list[i]))
        f.write('	   ! nelec  [cm^-3]   electron number density\n')
        f.write(fmt.format('{:.1m}',Tg_list[i]))
        f.write('	   ! Tgas [K]           gas temperature\n')
        f.write(fmt.format('{:.1m}',Td_list[i]))
        f.write('	   ! Tdust [K]           dust temperature\n')
        f.write(fmt.format('{:.1m}',vturb_list[i]))
        f.write('	   ! vturb  [kms/s]    turbulent velocity\n')
        f.write(fmt.format('{:.1m}',dust_to_gas_list[i]))
        f.write('	   ! dust_to_gas\n')
        f.write(f'{len(species_list)}	   ! Nspecies            number of species for that model\n')
        for j in range(len(species_list)):
            f.write(f'{list(species_list.keys())[j]}      {species_list[list(species_list.keys())[j]]}\n')
        f.write('           ! end\n')
        f.write('-------------------------------------------------------')
    f.close()
    return


class MyFormatter(string.Formatter):
    """
    String formatter for producing input files
    """

    def format_field(self,value,format_spec):
        if format_spec[-1]=='m':
            return super().format_field(value,format_spec[:-1]+'e').replace('e+','e')
        else:
            return super().format_field(value,format_spec)


fmt=MyFormatter()

